"""Format typehints elegantly and and fix automatically created links.

The Sphinx extension :mod:`sphinx_autodoc_typehints` adds type annotations to functions.
This extension modifies the created type annotations in four ways:

#. It formats the annotations more simply and in line with e.g. :mod:`numpy`.
#. It defines a configuration value ``qualname_overrides`` for ``conf.py``
   that overrides automatically created links. It is used like this::

       qualname_overrides = {
           "pandas.core.frame.DataFrame": "pandas.DataFrame",
           ...,
       }

   The defaults include :class:`anndata.AnnData`, :class:`pandas.DataFrame`,
   :class:`scipy.sparse.spmatrix` and other classes in :mod:`scipy.sparse`.

   It is necessary since :attr:`~definition.__qualname__` does not necessarily match
   the documented location of the function/class.

   Once either `sphinx issue 4826`_ or `sphinx-autodoc-typehints issue 38`_ are fixed,
   this part of the functionality will no longer be necessary.
#. The config value ``annotate_defaults`` (default: :data:`True`) controls if rST code
   like ``(default: `42`)`` is added after the type.
   It sets sphinx-autodoc-typehints’s option ``typehints_defaults`` to ``'braces'``
#. Type annotations for :class:`tuple` return types are added::

       def x() -> Tuple[int, float]:
           \"""
           Returns:
               a: An integer
               b: A floating point number
           \"""

   will render as:

       :Returns: a : :class:`int`
                     An integer
                 b : :class:`float`
                     A floating point number


.. _sphinx issue 4826: https://github.com/sphinx-doc/sphinx/issues/4826
.. _sphinx-autodoc-typehints issue 38: https://github.com/agronholm/sphinx-autodoc-typehints/issues/38

"""
from __future__ import annotations

from collections import ChainMap
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Any

from docutils.parsers.rst import roles
from sphinx.application import Sphinx
from sphinx.config import Config
from sphinx.ext.autodoc import ClassDocumenter

from .. import _setup_sig, metadata
from .example import example_func


__all__ = ["example_func", "setup"]


HERE = Path(__file__).parent.resolve()

qualname_overrides_default = {
    "anndata.base.AnnData": "anndata.AnnData",
    "anndata.core.anndata.AnnData": "anndata.AnnData",
    "anndata._core.anndata.AnnData": "anndata.AnnData",
    "matplotlib.axes._axes.Axes": "matplotlib.axes.Axes",
    "pandas.core.frame.DataFrame": "pandas.DataFrame",
    "pandas.core.indexes.base.Index": "pandas.Index",
    "scipy.sparse.base.spmatrix": "scipy.sparse.spmatrix",
    "scipy.sparse.csr.csr_matrix": "scipy.sparse.csr_matrix",
    "scipy.sparse.csc.csc_matrix": "scipy.sparse.csc_matrix",
}
qualname_overrides = ChainMap({}, qualname_overrides_default)


def _init_vars(app: Sphinx, config: Config):
    qualname_overrides.update(config.qualname_overrides)
    if "sphinx_autodoc_typehints" not in config.extensions and config.annotate_defaults:
        raise ValueError(
            "Can only use annotate_defaults option when using sphinx-autodoc-typehints"
        )
    if config.typehints_defaults is None and config.annotate_defaults:
        # override default for “typehints_defaults”
        config.typehints_defaults = "braces"
    config.html_static_path.append(str(HERE / "static"))


@dataclass
class PickleableCallable:
    func: Callable

    __call__ = property(lambda self: self.func)


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Patches :mod:`sphinx_autodoc_typehints` for a more elegant display."""
    # TODO: Unsure if “html” is sufficient or if we need to do “env”;
    #       Depends on when the autodoc-process-docstring event is handled.
    app.add_config_value("qualname_overrides", {}, "html")
    app.add_config_value("annotate_defaults", True, "html")
    app.add_css_file("typehints.css")
    app.connect("config-inited", _init_vars)

    from .formatting import _role_annot, format_annotation

    app.config["typehints_formatter"] = PickleableCallable(format_annotation)
    for name in ["annotation-terse", "annotation-full"]:
        roles.register_canonical_role(
            name, partial(_role_annot, additional_classes=name.split("-"))
        )

    from .autodoc_patch import dir_head_adder

    ClassDocumenter.add_directive_header = dir_head_adder(
        qualname_overrides, ClassDocumenter.add_directive_header
    )

    from .return_tuple import process_docstring  # , process_signature

    app.connect("autodoc-process-docstring", process_docstring)
    # app.connect("autodoc-process-signature", process_signature)

    return metadata



"""GitHub URLs for class and method pages.

This extension does two things:

#. It registers a :ref:`Jinja filter <jinja:filters>` called :func:`github_url`
   that you can use to convert a module path into a GitHub URL.
#. It configures :mod:`sphinx.ext.linkcode` for you if loaded after it in ``conf.py``:

   .. code:: python

      extensions = [
          "scanpydoc",
          "sphinx.ext.linkcode",
      ]

      # no need to define `linkcode_resolve`

Configuration
-------------

Uses the following config values in ``conf.py``::

    rtd_links_prefix: os.PathLike | str = ...  # default: '.'

    # sphinx book theme style
    html_theme_options = dict(
        repository_url=...,
        repository_branch=...,
    )
    # or RTD theme style:
    html_context = dict(
        github_user=...,
        github_repo=...,
        github_version=...,
    )

The ``rtd_links_prefix`` is for figuring out the .py file path relative to the git root,
that is to construct the path in the GitHub URL.

Which html configuration style you want to use depends on your theme, e.g.
:doc:`Sphinx Book Theme <sphinx_book_theme:index>`.

``:github_url:`` usage
----------------------

You can use the filter e.g. in `autosummary templates`_.
To configure the :doc:`Sphinx Book Theme <sphinx_book_theme:index>`,
override the ``autosummary/base.rst`` template like this:

.. code:: restructuredtext

    :github_url: {{ fullname | github_url }}

    {% extends "!autosummary/base.rst" %}

.. _autosummary templates: \
   http://www.sphinx-doc.org/en/master/usage/extensions/autosummary.html#customizing-templates
"""
from __future__ import annotations

import inspect
import sys
from importlib import import_module
from pathlib import Path, PurePosixPath
from types import ModuleType
from typing import Any

from jinja2.defaults import DEFAULT_FILTERS
from sphinx.application import Sphinx
from sphinx.config import Config

from .. import _setup_sig, metadata


rtd_links_prefix: Path = None
github_base_url: str = None


def _init_vars(app: Sphinx, config: Config):
    """Called when ``conf.py`` has been loaded."""
    global github_base_url, rtd_links_prefix
    _check_html_config(config)
    try:
        github_base_url = "https://github.com/{github_user}/{github_repo}/tree/{github_version}".format_map(
            config.html_context
        )
    except KeyError:
        github_base_url = "{repository_url}/tree/{repository_branch}".format_map(
            config.html_theme_options
        )
    rtd_links_prefix = PurePosixPath(config.rtd_links_prefix)


def _get_obj_module(qualname: str) -> tuple[Any, ModuleType]:
    """Get a module/class/attribute and its original module by qualname."""
    modname = qualname
    attr_path = []
    while modname not in sys.modules:
        modname, leaf = modname.rsplit(".", 1)
        attr_path.insert(0, leaf)

    # retrieve object and find original module name
    mod = sys.modules[modname]
    obj = None
    del modname
    for attr_name in attr_path:
        try:
            thing = getattr(mod if obj is None else obj, attr_name)
        except AttributeError as e:
            if is_dataclass(obj):
                thing = next(f for f in fields(obj) if f.name == attr_name)
            else:
                try:
                    thing = import_module(f"{mod.__name__}.{attr_name}")
                except ImportError:
                    raise e from None
        if isinstance(thing, ModuleType):
            mod = thing
        else:
            obj = thing
            mod_orig = getattr(obj, "__module__", None)
            if mod_orig is not None:
                mod = sys.modules[mod_orig]

    return obj, mod


def _get_linenos(obj):
    """Get an object’s line numbers."""
    try:
        lines, start = inspect.getsourcelines(obj)
    except TypeError:
        return None, None
    else:
        return start, start + len(lines) - 1


def _module_path(obj: Any, module: ModuleType) -> PurePosixPath:
    """Relative module path to parent directory of toplevel module."""
    try:
        file = Path(inspect.getabsfile(obj))
    except TypeError:
        file = Path(module.__file__)
    offset = -1 if file.name == "__init__.py" else 0
    parts = module.__name__.split(".")
    return PurePosixPath(*file.parts[offset - len(parts) :])


def github_url(qualname: str) -> str:
    """Get the full GitHub URL for some object’s qualname.

    Args:
        qualname: The full qualified name of a function, class, method or module

    Returns:
        A GitHub URL derived from the :confval:`html_context`
        or the :confval:`html_theme_options`.
    """
    try:
        obj, module = _get_obj_module(qualname)
    except Exception:
        print(f"Error in github_url({qualname!r}):", file=sys.stderr)
        raise
    path = rtd_links_prefix / _module_path(obj, module)
    start, end = _get_linenos(obj)
    fragment = f"#L{start}-L{end}" if start and end else ""
    return f"{github_base_url}/{path}{fragment}"


def _check_html_config(config: Config):
    options = dict(
        html_context={"github_user", "github_repo", "github_version"},
        html_theme_options={"repository_url", "repository_branch"},
    )
    if not any(name in config for name in options):
        msg = (
            f"Extension {__name__} needs “html_context” "
            "or “html_theme_options” to be defined in conf.py"
        )
        raise ValueError(msg)
    missing_value_sets = {
        name: opts - config[name].keys() for name, opts in options.items()
    }
    if not all(missing_value_sets.values()):
        return  # a valid configuration was found

    mvs = " or ".join(
        f"{name} {', '.join(repr(mv) for mv in mvs)}"
        for name, mvs in missing_value_sets.items()
    )
    msg = f"Extension {__name__} needs {mvs} to be defined in conf.py."
    for name in options:
        if name in config:
            msg += f"\n{name} = {config[name]!r}"
    raise ValueError(msg)


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Register the :func:`github_url` :ref:`Jinja filter <jinja:filters>`."""

    app.add_config_value("rtd_links_prefix", PurePosixPath("."), "")
    app.connect("config-inited", _init_vars)

    # if linkcode config not set
    if "linkcode_resolve" not in app.config or app.config["linkcode_resolve"] is None:
        from ._linkcode import linkcode_resolve

        app.config["linkcode_resolve"] = linkcode_resolve

    # html_context doesn’t apply to autosummary templates ☹
    # and there’s no way to insert filters into those templates
    # so we have to modify the default filters
    DEFAULT_FILTERS["github_url"] = github_url

    return metadata


if True:  # test data
    from dataclasses import dataclass, field, fields, is_dataclass

    @dataclass
    class _TestCls:
        test_attr: dict[str, str] = field(default_factory=dict)



from __future__ import annotations


def example_func(a: str | None, b: str | int | None = None) -> dict[str, int]:
    """Example function.

    Hover over the parameter and return type annotations to see the long versions.

    Args:
        a: An example parameter
        b: Another, with a default
    Returns:
        An example return value
    """
    pass



from __future__ import annotations

import inspect
from collections.abc import Callable, Iterable, Mapping, Sequence
from functools import partial
from typing import Any, Literal, get_args, get_origin

from docutils import nodes
from docutils.nodes import Node
from docutils.parsers.rst.roles import set_classes
from docutils.parsers.rst.states import Inliner, Struct
from docutils.utils import SystemMessage, unescape
from sphinx.config import Config
from sphinx_autodoc_typehints import format_annotation as _format_orig

from scanpydoc import elegant_typehints


def _format_full(annotation: type[Any], config: Config) -> str | None:
    if inspect.isclass(annotation) and annotation.__module__ == "builtins":
        return None

    origin = get_origin(annotation)
    tilde = "" if config.typehints_fully_qualified else "~"

    annotation_cls = annotation if inspect.isclass(annotation) else type(annotation)
    if annotation_cls.__module__ == "typing":
        return None

    # Only if this is a real class we override sphinx_autodoc_typehints
    if inspect.isclass(annotation) or inspect.isclass(origin):
        try:
            full_name = f"{annotation.__module__}.{annotation.__qualname__}"
        except AttributeError:
            full_name = f"{origin.__module__}.{origin.__qualname__}"
        override = elegant_typehints.qualname_overrides.get(full_name)
        role = "exc" if issubclass(annotation_cls, BaseException) else "class"
        if override is not None:
            return f":py:{role}:`{tilde}{override}`"

    return None


def _format_terse(annotation: type[Any], config: Config) -> str:
    from typing import Mapping as t_Mapping
    from typing import Union

    origin = get_origin(annotation)
    args = get_args(annotation)
    tilde = "" if config.typehints_fully_qualified else "~"
    fmt = partial(_format_terse, config=config)

    # display `Union[A, B]` as `A | B`
    if origin is Union:
        # Never use the `Optional` keyword in the displayed docs.
        # Use `| None` instead, similar to other large numerical packages.
        return " | ".join(map(fmt, args))

    # do not show the arguments of Mapping
    if origin is Mapping or origin is t_Mapping:
        return f":py:class:`{tilde}collections.abc.Mapping`"

    # display dict as {k: v}
    if origin is dict and len(args) == 2:
        k, v = args
        return f"{{{fmt(k)}: {fmt(v)}}}"

    # display Callable[[a1, a2], r] as (a1, a2) -> r
    if origin is Callable and len(args) == 2:
        params, ret = args
        params = ["…"] if params is Ellipsis else map(fmt, params)
        return f"({', '.join(params)}) → {fmt(ret)}"

    # display Literal as {'a', 'b', ...}
    if origin is Literal:
        return f"{{{', '.join(map(repr, args))}}}"

    return _format_full(annotation, config) or _format_orig(annotation, config)


def format_annotation(annotation: type[Any], config: Config) -> str | None:
    """Generate reStructuredText containing links to the types.

    Unlike :func:`sphinx_autodoc_typehints.format_annotation`,
    it tries to achieve a simpler style as seen in numeric packages like numpy.

    Args:
        annotation: A type or class used as type annotation.
        config: Sphinx config containing ``sphinx-autodoc-typehints``’s options.

    Returns:
        reStructuredText describing the type
    """
    curframe = inspect.currentframe()
    calframe = inspect.getouterframes(curframe, 2)
    if calframe[2].function in {"process_docstring", "_inject_signature"} or (
        calframe[2].function == "_inject_types_to_docstring"
        and calframe[3].function == "process_docstring"
    ):
        return format_both(annotation, config)
    else:  # recursive use
        return _format_full(annotation, config)


def format_both(annotation: type[Any], config: Config) -> str:
    terse = _format_terse(annotation, config)
    full = _format_full(annotation, config) or _format_orig(annotation, config)
    if terse == full:
        return terse
    return f":annotation-terse:`{_escape(terse)}`\\ :annotation-full:`{_escape(full)}`"


def _role_annot(
    name: str,
    rawtext: str,
    text: str,
    lineno: int,
    inliner: Inliner,
    options: dict[str, Any] = {},
    content: Sequence[str] = (),
    # *,  # https://github.com/ambv/black/issues/613
    additional_classes: Iterable[str] = (),
) -> tuple[list[Node], list[SystemMessage]]:
    options = options.copy()
    set_classes(options)
    if additional_classes:
        options["classes"] = options.get("classes", []).copy()
        options["classes"].extend(additional_classes)
    memo = Struct(
        document=inliner.document, reporter=inliner.reporter, language=inliner.language
    )
    node = nodes.inline(unescape(rawtext), "", **options)
    children, messages = inliner.parse(_unescape(text), lineno, memo, node)
    node.extend(children)
    return [node], messages


def _escape(rst: str) -> str:
    return rst.replace("`", "\\`")


def _unescape(rst: str) -> str:
    # TODO: IDK why the [ part is necessary.
    return unescape(rst).replace("\\`", "`").replace("[", "\\[")



from datetime import datetime
from importlib.metadata import metadata
from pathlib import PurePosixPath

from sphinx.application import Sphinx


extensions = [
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",  # needs to be after napoleon
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "scanpydoc",
    "sphinx.ext.linkcode",  # needs to be after scanpydoc
]

intersphinx_mapping = dict(
    python=("https://docs.python.org/3", None),
    jinja=("https://jinja.palletsprojects.com/en/2.10.x/", None),
    sphinx=("https://www.sphinx-doc.org/en/master/", None),
    sphinx_book_theme=("https://sphinx-book-theme.readthedocs.io/en/stable/", None),
    # examples
    numpy=("https://numpy.org/doc/stable/", None),
    anndata=("https://anndata.readthedocs.io/en/latest/", None),
    pandas=("https://pandas.pydata.org/pandas-docs/stable/", None),
    scipy=("https://docs.scipy.org/doc/scipy/", None),
)

# general information
meta = metadata("scanpydoc")
project = meta["name"]
author = meta["author-email"].split(" <")[0]
copyright = f"{datetime.now():%Y}, {author}."
version = release = meta["version"]

master_doc = "index"
templates_path = ["_templates"]

# Generate .rst stubs for modules using autosummary
autosummary_generate = True
autosummary_ignore_module_all = False
# Don’t add module paths to documented functions’ names
add_module_names = False

html_theme = "scanpydoc"
html_theme_options = dict(
    repository_url="https://github.com/theislab/scanpydoc",
    repository_branch="main",
    use_repository_button=True,
)

rtd_links_prefix = PurePosixPath("src")


def setup(app: Sphinx):
    app.add_object_type(
        "confval",
        "confval",
        objname="configuration value",
        indextemplate="pair: %s; configuration value",
    )



from __future__ import annotations

import importlib.util
import linecache
import sys
from collections.abc import Callable
from textwrap import dedent
from uuid import uuid4

import pytest
from docutils.nodes import document
from sphinx.application import Sphinx
from sphinx.testing.path import path as STP


@pytest.fixture
def make_app_setup(make_app, tmp_path) -> Callable[..., Sphinx]:
    def make_app_setup(**conf) -> Sphinx:
        (tmp_path / "conf.py").write_text("")
        return make_app(srcdir=STP(tmp_path), confoverrides=conf)

    return make_app_setup


@pytest.fixture
def render() -> Callable[[Sphinx, document], str]:
    def _render(app: Sphinx, doc: document) -> str:
        # Doesn’t work as desc is an Admonition and the HTML writer doesn’t handle it
        # print(app.builder.render_partial(doc[1]))
        app.builder.prepare_writing({doc["source"]})
        writer = app.builder.docwriter  # type: Writer
        writer.document = doc
        writer.document.settings = app.builder.docsettings
        writer.translate()
        return writer.output

    return _render


@pytest.fixture
def make_module(tmp_path):
    added_modules = []

    def make_module(name, code):
        code = dedent(code)
        assert name not in sys.modules
        spec = importlib.util.spec_from_loader(name, loader=None)
        mod = sys.modules[name] = importlib.util.module_from_spec(spec)
        path = tmp_path / f"{name}_{str(uuid4()).replace('-', '_')}.py"
        path.write_text(code)
        mod.__file__ = str(path)
        exec(code, mod.__dict__)
        linecache.updatecache(str(path), mod.__dict__)
        added_modules.append(name)
        return mod

    yield make_module

    for name in added_modules:
        del sys.modules[name]



"""Prettier function parameter documentation.

This extension replaces the default :class:`~sphinx.domains.python.PyTypedField`
with a derivative :class:`DLTypedField`, which renders item items
(e.g. function parameters) as definition lists instead of simple paragraphs.
"""

from __future__ import annotations

from typing import Any

from docutils import nodes
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.domains.python import PyObject, PyTypedField
from sphinx.environment import BuildEnvironment

from . import _setup_sig, metadata


class DLTypedField(PyTypedField):
    """A reStructuredText field-list renderer that creates definition lists.

    This style is more readable than cramming name, type, and description in one line.

    The class is not intended to be used directly, but by using the extension,
    it will be used instead of the default :class:`~sphinx.domains.python.PyTypedField`.
    """

    #: Override the list type
    list_type = nodes.definition_list

    def make_field(
        self,
        types: dict[str, list[nodes.Node]],
        domain: str,
        items: tuple[str, list[nodes.inline]],
        env: BuildEnvironment | None = None,
        **kw,
    ) -> nodes.field:
        """
        Render a field to a document-tree node representing a definition list item.
        """

        def make_refs(role_name, name, node):
            return self.make_xrefs(role_name, domain, name, node, env=env, **kw)

        def handle_item(
            fieldarg: str, content: list[nodes.inline]
        ) -> nodes.definition_list_item:
            term = nodes.term()
            term += make_refs(self.rolename, fieldarg, addnodes.literal_strong)

            field_type = types.pop(fieldarg, None)
            if field_type is not None:
                if len(field_type) == 1 and isinstance(field_type[0], nodes.Text):
                    (text_node,) = field_type  # type: nodes.Text
                    classifier_content = make_refs(
                        self.typerolename, text_node.astext(), addnodes.literal_emphasis
                    )
                else:
                    classifier_content = field_type
                term += [
                    # https://github.com/sphinx-doc/sphinx/issues/10815
                    nodes.Text(" "),
                    # Sphinx tries to fixup classifiers without rawsource,
                    # but for this expects attributes we don’t have. Thus “×”.
                    nodes.classifier("×", "", *classifier_content),
                ]

            def_content = nodes.paragraph("", "", *content)
            definition = nodes.definition("", def_content)

            return nodes.definition_list_item("", term, definition)

        field_name = nodes.field_name("", self.label)
        assert not self.can_collapse
        body_node = self.list_type(classes=["simple"])
        for field_arg, content in items:
            body_node += handle_item(field_arg, content)
        field_body = nodes.field_body("", body_node)
        return nodes.field("", field_name, field_body)


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Replace :class:`~sphinx.domains.python.PyTypedField` with ours."""
    napoleon_requested = "sphinx.ext.napoleon" in app.config.extensions
    napoleon_loaded = next(
        (True for ft in PyObject.doc_field_types if ft.name == "keyword"), False
    )
    if napoleon_requested and not napoleon_loaded:
        raise RuntimeError(f"Please load sphinx.ext.napoleon before {__name__}")

    PyObject.doc_field_types = [
        DLTypedField(
            ft.name,
            names=ft.names,
            typenames=ft.typenames,
            label=ft.label,
            rolename=ft.rolename,
            typerolename=ft.typerolename,
            # Definition lists can’t collapse.
            can_collapse=False,
        )
        if isinstance(ft, PyTypedField)
        else ft
        for ft in PyObject.doc_field_types
    ]

    return metadata



"""A series of Sphinx extensions to get maintainable numpydoc style documentation.

This module is also an extension itself which simply sets up all included extensions.
"""
from __future__ import annotations

from textwrap import indent
from typing import Any

from sphinx.application import Sphinx

from ._version import __version__


metadata = dict(version=__version__, env_version=1, parallel_read_safe=True)

# Can’t seem to be able to do this in numpydoc style:
# https://github.com/sphinx-doc/sphinx/issues/5887
setup_sig_str = """\
Args:
    app: Sphinx app to set this :term:`sphinx:extension` up for

Returns:
    :ref:`Metadata <sphinx:ext-metadata>` for this extension.
"""


def _setup_sig(fn):
    fn.__doc__ += "\n\n" + indent(setup_sig_str, " " * 4)
    return fn


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Set up all included extensions!."""
    app.setup_extension("scanpydoc.autosummary_generate_imported")
    app.setup_extension("scanpydoc.definition_list_typed_field")
    app.setup_extension("scanpydoc.elegant_typehints")
    app.setup_extension("scanpydoc.rtd_github_links")
    app.setup_extension("scanpydoc.theme")
    return metadata



from __future__ import annotations

import inspect
import re
import sys
import typing as t
from collections.abc import Callable, Mapping
from pathlib import Path
from typing import Any

import pytest
import sphinx_autodoc_typehints as sat
from sphinx.application import Sphinx
from sphinx.testing.restructuredtext import parse

from scanpydoc.elegant_typehints.formatting import (
    _format_full,
    _format_terse,
    format_annotation,
)
from scanpydoc.elegant_typehints.return_tuple import process_docstring


@pytest.fixture
def _testmod(make_module):
    return make_module(
        "_testmod",
        """\
        class Class: pass
        class SubCl(Class): pass
        class Excep(RuntimeError): pass
        class Excep2(Excep): pass
        """,
    )


@pytest.fixture
def app(make_app_setup) -> Sphinx:
    return make_app_setup(
        master_doc="index",
        extensions=[
            "sphinx.ext.autodoc",
            "sphinx.ext.napoleon",
            "sphinx_autodoc_typehints",
            "scanpydoc.elegant_typehints",
        ],
        qualname_overrides={
            "_testmod.Class": "test.Class",
            "_testmod.SubCl": "test.SubCl",
            "_testmod.Excep": "test.Excep",
            "_testmod.Excep2": "test.Excep2",
        },
    )


@pytest.fixture
def process_doc(app):
    def process(fn: Callable) -> list[str]:
        lines = inspect.getdoc(fn).split("\n")
        sat.process_docstring(app, "function", fn.__name__, fn, None, lines)
        process_docstring(app, "function", fn.__name__, fn, None, lines)
        return lines

    return process


def test_app(app):
    assert "qualname_overrides" in app.config.values
    assert "_testmod.Class" in app.config.qualname_overrides


def test_default(app):
    assert format_annotation(str, app.config) is None


def test_alternatives(process_doc):
    def fn_test(s: str):
        """:param s: Test"""

    assert process_doc(fn_test) == [":type s: :py:class:`str`", ":param s: Test"]


def test_defaults_simple(process_doc):
    def fn_test(s: str = "foo", n: None = None, i_: int = 1):
        r""":param s: Test S
        :param n: Test N
        :param i\_: Test I
        """

    assert process_doc(fn_test) == [
        ":type s: :py:class:`str` (default: ``'foo'``)",
        ":param s: Test S",
        ":type n: :py:obj:`None` (default: ``None``)",
        ":param n: Test N",
        r":type i\_: :py:class:`int` (default: ``1``)",
        r":param i\_: Test I",
    ]


@pytest.mark.skipif(sys.version_info < (3, 9), reason="s-a-t bug in Python 3.8")
def test_defaults_complex(process_doc):
    def fn_test(m: Mapping[str, int] = {}):
        """:param m: Test M"""

    assert process_doc(fn_test) == [
        ":type m: "
        r":annotation-terse:`:py:class:\`~collections.abc.Mapping\``\ "
        r":annotation-full:`"
        r":py:class:\`~collections.abc.Mapping\`\[:py:class:\`str\`, :py:class:\`int\`]"
        "` (default: ``{}``)",
        ":param m: Test M",
    ]


def test_mapping(app):
    assert (
        _format_terse(t.Mapping[str, Any], app.config)
        == ":py:class:`~collections.abc.Mapping`"
    )
    assert _format_full(t.Mapping[str, Any], app.config) is None


def test_dict(app):
    assert _format_terse(t.Dict[str, Any], app.config) == (
        "{:py:class:`str`: :py:data:`~typing.Any`}"
    )


@pytest.mark.parametrize(
    "annotation,expected",
    [
        (t.Callable[..., t.Any], "(…) → :py:data:`~typing.Any`"),
        (
            t.Callable[[str, int], None],
            "(:py:class:`str`, :py:class:`int`) → :py:obj:`None`",
        ),
    ],
)
def test_callable_terse(app, annotation, expected):
    assert _format_terse(annotation, app.config) == expected


def test_literal(app):
    assert _format_terse(t.Literal["str", 1, None], app.config) == "{'str', 1, None}"
    assert _format_full(t.Literal["str", 1, None], app.config) is None


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Syntax only available on 3.10+")
def test_syntax_vs_typing(app):
    u = eval("int | str")
    assert _format_terse(u, app.config) == ":py:class:`int` | :py:class:`str`"
    assert _format_full(u, app.config) is None  # this used to crash


def test_qualname_overrides_class(app, _testmod):
    assert _testmod.Class.__module__ == "_testmod"
    assert _format_terse(_testmod.Class, app.config) == ":py:class:`~test.Class`"


def test_qualname_overrides_exception(app, _testmod):
    assert _testmod.Excep.__module__ == "_testmod"
    assert _format_terse(_testmod.Excep, app.config) == ":py:exc:`~test.Excep`"


def test_qualname_overrides_recursive(app, _testmod):
    assert _format_terse(t.Union[_testmod.Class, str], app.config) == (
        r":py:class:`~test.Class` | :py:class:`str`"
    )
    assert _format_full(t.Union[_testmod.Class, str], app.config) is None


def test_fully_qualified(app, _testmod):
    app.config.typehints_fully_qualified = True
    assert _format_terse(t.Union[_testmod.Class, str], app.config) == (
        r":py:class:`test.Class` | :py:class:`str`"
    )
    assert _format_full(t.Union[_testmod.Class, str], app.config) is None


def test_classes_get_added(app):
    doc = parse(app, r":annotation-full:`:py:class:\`str\``")
    assert doc[0].tagname == "paragraph"
    assert doc[0][0].tagname == "inline"
    assert doc[0][0]["classes"] == ["annotation", "full"]
    # print(doc.asdom().toprettyxml())


@pytest.mark.parametrize("formatter", [_format_terse, _format_full], ids="tf")
# These guys aren’t listed as classes in Python’s intersphinx index:
@pytest.mark.parametrize(
    "annotation",
    [
        t.Any,
        t.AnyStr,
        # t.NoReturn,
        t.Callable[[int], None],
        # t.ClassVar[t.Any],
        t.Tuple[int, str],
        t.Tuple[float, ...],
        t.Union[int, str],
        t.Union[int, str, None],
    ],
    ids=lambda p: str(p).replace("typing.", ""),
)
def test_typing_classes(app, annotation, formatter):
    app.config.typehints_fully_qualified = True
    name = (
        getattr(annotation, "_name", None)
        or getattr(annotation, "__name__", None)
        or getattr(t.get_origin(annotation), "_name", None)
        # 3.6 _Any and _Union
        or annotation.__class__.__name__[1:]
    )
    if formatter is _format_terse and name in {"Union", "Callable"}:
        pytest.skip("Tested elsewhere")
    output = formatter(annotation, app.config)
    assert output is None or output.startswith(f":py:data:`typing.{name}")


@pytest.mark.parametrize(
    "direc,base,sub",
    [
        ("autoclass", "Class", "SubCl"),
        ("autoexception", "Excep", "Excep2"),
    ],
)
def test_autodoc(app, _testmod, direc, base, sub):
    Path(app.srcdir, "index.rst").write_text(
        f"""\
.. {direc}:: _testmod.{sub}
   :show-inheritance:
"""
    )
    app.build()
    out = Path(app.outdir, "index.html").read_text()
    assert not app._warning.getvalue(), app._warning.getvalue()
    assert re.search(
        r'<(code|span)?[^>]*><span class="pre">test\.</span></(code|span)>'
        f'<(code|span)?[^>]*><span class="pre">{sub}</span></(code|span)>',
        out,
    ), out
    assert f'<a class="headerlink" href="#test.{sub}"' in out, out
    assert re.search(rf"Bases: <code[^>]*><span[^>]*>(?:test\.)?{base}", out), out


def test_fwd_ref(app, make_module):
    make_module(
        "fwd_mod",
        """\
        from dataclasses import dataclass

        @dataclass
        class A:
           b: 'B'

        @dataclass
        class B:
            a: A
        """,
    )
    Path(app.srcdir, "index.rst").write_text(
        """\
.. autosummary::

   fwd_mod.A
   fwd_mod.B
"""
    )
    app.setup_extension("sphinx.ext.autosummary")

    app.build()

    out = Path(app.outdir, "index.html").read_text()
    warnings = [
        w
        for w in app._warning.getvalue().splitlines()
        if "Cannot treat a function defined as a local function" not in w
    ]
    assert not warnings, warnings
    # TODO: actually reproduce #14
    assert "fwd_mod.A" in out, out


@pytest.mark.parametrize(
    "docstring",
    [
        """
        :param: x
        :return: foo
                     A foo!
                 bar
                     A bar!
        """,
        """
        :return: foo
                     A foo!
                 bar
                     A bar!
        :param: x
        """,
    ],
    ids=["Last", "First"],
)
@pytest.mark.parametrize(
    "return_ann, foo_rendered",
    [
        (t.Tuple[str, int], ":py:class:`str`"),
        (t.Optional[t.Tuple[str, int]], ":py:class:`str`"),
        (
            t.Tuple[t.Mapping[str, float], int],
            r":annotation-terse:`:py:class:\`~collections.abc.Mapping\``\ "
            r":annotation-full:`:py:class:\`~typing.Mapping\`\["
            r":py:class:\`str\`, :py:class:\`float\`"
            r"]`",
        ),
    ],
    ids=["Tuple", "Optional[Tuple]", "Complex"],
)
def test_return(process_doc, docstring, return_ann, foo_rendered):
    def fn_test():
        pass

    fn_test.__doc__ = docstring
    fn_test.__annotations__["return"] = return_ann
    lines = [
        l
        for l in process_doc(fn_test)
        if not re.match("^:(rtype|param|annotation-(full|terse)):", l)
    ]
    assert lines == [
        f":return: foo : {foo_rendered}",
        "             A foo!",
        "         bar : :py:class:`int`",
        "             A bar!",
    ]


def test_return_too_many(process_doc):
    def fn_test() -> tuple[int, str]:
        """:return: foo
        A foo!
        bar
        A bar!
        baz
        A baz!
        """

    assert not any(
        "annotation-terse" in l
        for l in process_doc(fn_test)
        if not l.startswith(":rtype:")
    )



from dataclasses import Field
from importlib import import_module
from pathlib import Path, PurePosixPath

import pytest
from _pytest.monkeypatch import MonkeyPatch

from scanpydoc.rtd_github_links import _get_linenos, _get_obj_module, github_url


HERE = Path(__file__).parent


@pytest.fixture
def _env(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr("scanpydoc.rtd_github_links.github_base_url", "x")


@pytest.fixture(params=[".", "src"])
def prefix(monkeypatch: MonkeyPatch, _env, request) -> PurePosixPath:
    pfx = PurePosixPath(request.param)
    monkeypatch.setattr("scanpydoc.rtd_github_links.rtd_links_prefix", pfx)
    return "x" / pfx / "scanpydoc"


@pytest.mark.parametrize(
    ("module", "name", "obj_path"),
    [
        pytest.param(
            *("rtd_github_links", "github_url", "rtd_github_links/__init__.py"),
            id="basic",
        ),
        pytest.param(
            *("elegant_typehints", "example_func", "elegant_typehints/example.py"),
            id="reexport",
        ),
    ],
)
def test_as_function(prefix, module, name, obj_path):
    assert github_url(f"scanpydoc.{module}") == str(prefix / module / "__init__.py")
    obj = getattr(import_module(f"scanpydoc.{module}"), name)
    s, e = _get_linenos(obj)
    assert github_url(f"scanpydoc.{module}.{name}") == f"{prefix}/{obj_path}#L{s}-L{e}"


def test_get_obj_module():
    import sphinx.application as sa

    obj, mod = _get_obj_module("scanpydoc.Sphinx")
    assert obj is sa.Sphinx
    assert mod is sa


def test_get_obj_module_anntation():
    obj, mod = _get_obj_module("scanpydoc.rtd_github_links._TestCls.test_attr")
    assert isinstance(obj, Field)



"""A widescreen extension for :doc:`sphinx_book_theme:index`.

Add to ``conf.py``:

.. code:: python

   html_theme = 'scanpydoc'

Theme options
=============

This theme adds the following configuration option,
and the ones under `docsearch options`_:

.. confval:: accent_color

   :type: str
   :default: ``#f07e44``

   The CSS color used for the mobile header background and the project name text.

See ``sphinx_book_theme``’s :doc:`sphinx_book_theme:reference`, e.g.:

.. code:: python

   html_theme_options = dict(
       repository_url="https://github.com/theislab/scanpydoc",
       repository_branch="main",
       use_repository_button=True,
   )

Docsearch options
-----------------

These two configuration values are required to use docsearch_:

.. _docsearch: https://docsearch.algolia.com/

.. confval:: docsearch_key

   :type: str

   The API key provided by docsearch.

.. confval:: docsearch_index

   :type: str

   The index name used by docsearch.

The following configuration values are optional:

.. confval:: docsearch_doc_version

   :type: str
   :default: ``'latest'`` or ``'stable'``

   The documentation version searched.
   The default is ``'stable'`` if ``READTHEDOCS_VERSION=stable`` is set,
   and ``'latest'`` otherwise.

.. confval:: docsearch_js_version

   :type: str
   :default: ``'2.6'``

   The docsearch library version used.

"""

from pathlib import Path

from sphinx.application import Sphinx

from .. import _setup_sig


HERE = Path(__file__).parent.resolve()


@_setup_sig
def setup(app: Sphinx):
    """Setup theme (like an extension)."""
    app.add_html_theme("scanpydoc", str(HERE))
    return dict(parallel_read_safe=True, parallel_write_safe=True)



from __future__ import annotations

from typing import Literal, TypedDict


class PyInfo(TypedDict):
    module: str
    fullname: str


class CInfo(TypedDict):
    """C / C++ info."""

    names: list[str]


class JSInfo(TypedDict):
    object: str
    fullname: str


def linkcode_resolve(
    domain: Literal["py", "c", "cpp", "javascript"], info: PyInfo | CInfo | JSInfo
) -> str | None:
    from . import github_url

    if domain != "py":
        return None
    if not info["module"]:
        return None
    return github_url(f'{info["module"]}.{info["fullname"]}')



import pytest
from sphinx.application import Sphinx
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.testing.restructuredtext import parse
from sphinx.writers.html import HTMLTranslator
from sphinx.writers.html5 import HTML5Translator


@pytest.fixture
def app(make_app_setup) -> Sphinx:
    app = make_app_setup()
    app.setup_extension("scanpydoc.definition_list_typed_field")
    return app


# Avoid :class: to not get pending_xref. TODO: fix
params_code = """\
.. py:function:: test(a, b=None)

   :param a: First parameter
   :type a: ``str``
   :param b: Second parameter
   :type v: ``~typing.Optional``\\[``str``]
"""

params_code_single = """\
.. py:function:: test(a)

   :param a: First parameter
   :type a: ``str``
"""


def test_apps_separate(app, make_app_setup):
    app_no_setup = make_app_setup()
    assert app is not make_app_setup
    assert "scanpydoc.definition_list_typed_field" in app.extensions
    assert "scanpydoc.definition_list_typed_field" not in app_no_setup.extensions


@pytest.mark.parametrize("code,n", [(params_code, 2), (params_code_single, 1)])
def test_convert_params(app, code, n):
    # the directive class is PyModuleLevel → PyObject → ObjectDescription
    # ObjectDescription.run uses a DocFieldTransformer to transform members
    # the signature of each Directive(
    #     name, arguments, options, content, lineno,
    #     content_offset, block_text, state, state_machine,
    # )

    doc = parse(app, code)
    assert doc[1].tagname == "desc"
    assert doc[1]["desctype"] == "function"
    assert doc[1][1].tagname == "desc_content"
    assert doc[1][1][0].tagname == "field_list"
    assert doc[1][1][0][0].tagname == "field"
    assert doc[1][1][0][0][0].tagname == "field_name"
    assert doc[1][1][0][0][0][0].astext() == "Parameters"
    assert doc[1][1][0][0][1].tagname == "field_body"
    assert doc[1][1][0][0][1][0].tagname == "definition_list"

    dl = doc[1][1][0][0][1][0]
    assert len(dl) == n, dl
    for dli in dl:
        assert dli.tagname == "definition_list_item"
        assert dli[0].tagname == "term"
        assert dli[1].tagname == "definition"

    # print(doc.asdom().toprettyxml())


@pytest.mark.parametrize("translator", [HTMLTranslator, HTML5Translator])
@pytest.mark.parametrize("code", [params_code, params_code_single])
def test_render_params_html4(app, render, translator, code):
    app.config.html4_writer = translator is HTMLTranslator
    assert app.builder.__class__ is StandaloneHTMLBuilder
    assert app.builder.default_translator_class is translator

    doc = parse(app, code)
    html = render(app, doc)
    assert "<dl" in html



"""Generate autosummary docs for imported members.

This extension patches the :mod:`~sphinx.ext.autosummary` extension
to generate docs for imported members.
It needs to be loaded and :confval:`autosummary_generate` needs to be set to ``True``.

This will hopefully be superseded by the ability to add ``:imported-members:``
to `autosummary templates`_ in the future. See `Sphinx issue 4372`_.

.. _autosummary templates: http://www.sphinx-doc.org/en/master/usage/extensions/autosummary.html#customizing-templates
.. _Sphinx issue 4372: https://github.com/sphinx-doc/sphinx/issues/4372
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

from sphinx.application import Sphinx
from sphinx.ext import autosummary
from sphinx.ext.autosummary.generate import generate_autosummary_docs

from . import _setup_sig, metadata


logger = logging.getLogger(__name__)


def _generate_stubs(app: Sphinx):
    gen_files = app.config.autosummary_generate

    if gen_files and not hasattr(gen_files, "__len__"):
        env = app.builder.env
        gen_files = [
            env.doc2path(x, base=None)
            for x in env.found_docs
            if Path(env.doc2path(x)).is_file()
        ]
    if not gen_files:
        return

    ext = app.config.source_suffix
    gen_files = [
        genfile + ("" if genfile.endswith(tuple(ext)) else ext[0])
        for genfile in gen_files
    ]

    suffix = autosummary.get_rst_suffix(app)
    if suffix is None:
        return

    generate_autosummary_docs(
        gen_files,
        builder=app.builder,
        warn=logger.warning,
        info=logger.info,
        suffix=suffix,
        base_path=app.srcdir,
        imported_members=True,
        app=app,
    )


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Patch autosummary to generate docs for imported members as well."""
    autosummary.process_generate_options = _generate_stubs

    return metadata



from __future__ import annotations

import inspect
import re
from logging import getLogger
from typing import Any, get_args, get_origin, get_type_hints

from sphinx.application import Sphinx
from sphinx.ext.autodoc import Options

from .formatting import format_both


logger = getLogger(__name__)
re_ret = re.compile("^:returns?: ")


def get_tuple_annot(annotation: type | None) -> tuple[type, ...] | None:
    from typing import Tuple, Union

    if annotation is None:
        return None
    origin = get_origin(annotation)
    if not origin:
        return None
    if origin is Union:
        for annotation in get_args(annotation):
            origin = get_origin(annotation)
            if origin in (tuple, Tuple):
                break
        else:
            return None
    return get_args(annotation)


def process_docstring(
    app: Sphinx,
    what: str,
    name: str,
    obj: Any,
    options: Options | None,
    lines: list[str],
) -> None:
    # Handle complex objects
    if isinstance(obj, property):
        obj = obj.fget
    if not callable(obj):
        return
    if what in ("class", "exception"):
        obj = obj.__init__
    obj = inspect.unwrap(obj)
    try:
        hints = get_type_hints(obj)
    except (AttributeError, TypeError):
        # Introspecting a slot wrapper will raise TypeError
        return
    ret_types = get_tuple_annot(hints.get("return"))
    if ret_types is None:
        return

    # Get return section
    i_prefix = None
    l_start = None
    for l, line in enumerate(lines):
        if i_prefix is None:
            m = re_ret.match(line)
            if m:
                i_prefix = m.span()[1]
                l_start = l
        elif len(line[:i_prefix].strip()) > 0:
            l_end = l - 1
            break
    else:
        l_end = len(lines) - 1
    if i_prefix is None:
        return

    # Meat
    idxs_ret_names = []
    for l, line in enumerate([l[i_prefix:] for l in lines[l_start : l_end + 1]]):
        if line.isidentifier() and lines[l + l_start + 1].startswith("    "):
            idxs_ret_names.append(l + l_start)

    if len(idxs_ret_names) == len(ret_types):
        for l, rt in zip(idxs_ret_names, ret_types):
            typ = format_both(rt, app.config)
            lines[l : l + 1] = [f"{lines[l]} : {typ}"]
